{-# LANGUAGE OverloadedStrings #-}

module Main (
    main,
) where

import qualified Formula as Fm
import qualified Formula.Automata as Fm
import qualified Formula.Simplify as Fm
import qualified RRn
import qualified TA
import Util.Pretty
import Util.Rewriting
import Util.Signature

import qualified Problem as P
import qualified Data.Rewriting.Rule as R
import qualified Data.Rewriting.Term as T

import Data.Text.Prettyprint.Doc hiding (list)
import Data.Text.Prettyprint.Doc.Render.Text
import qualified Data.Text.Lazy as Text
import qualified Data.Text.Lazy.IO as Text

import Text.Parsec (parse)
import System.Environment
import System.Console.GetOpt
import System.Exit
import System.IO
import Control.Monad
import qualified Util.Set as S
import Data.Maybe

data Options = Options {
    optVerbose :: Bool,
    optWitness :: Bool,
    optCertificate :: Maybe Handle,
    optInfo :: Bool
}

opt0 = Options {
    optVerbose = False,
    optWitness = False,
    optCertificate = Nothing,
    optInfo = False
}

certOn :: Options -> Bool
certOn opt = case optCertificate opt of
  Nothing -> False
  Just _ -> True

options :: [OptDescr (Options -> IO Options)]
options = [
    Option "h?"  ["help"]
        (NoArg (\_ -> usage ExitSuccess))
        "print usage information",
    Option "v"  ["verbose"]
        (NoArg (\o -> return $ o{ optVerbose = True }))
        "verbose output",
    Option "w"  ["witness"]
        (NoArg (\o -> return $ o{ optWitness = True }))
        "enable witness generation",
    Option "c"  ["certificate"]
        (ReqArg (\file o -> do
                        fileHandle <- openFile file WriteMode
                        return $ o { optCertificate = Just fileHandle }
                ) "FILE"
        ) "write certificate to FILE",
    Option "i"  ["info"]
        (NoArg (\o -> return $ o{ optInfo = True }))
        "enable additional information in the certificate"
 ]

usage :: ExitCode -> IO a
usage e = do
    hPutStrLn stderr $
        usageInfo header options
    exitWith e
  where
    header = unlines [
        "Usage: fort-h [OPTION...] FORMULA FILE.trs..."
     ]

main :: IO ()
main = do
    args <- getArgs
    case getOpt Permute options args of
        (_, _, e : _) -> do
            hPutStrLn stderr $ "Error: " ++ e
            usage (ExitFailure 1)
        (os, fs, []) -> do
            opt <- foldl (>>=) (return opt0) os
            main' opt fs
  where
    main' opt (p : fs@(_ : _)) = case parse Fm.fortFormula "<arg>" p of
        Left e -> die (show e)
        Right f -> do
            when (optVerbose opt) $ do
                comment opt $ "input formula:" `nl` indent 4 (Fm.pretty f)
            let f' = Fm.exForm . Fm.unrestrict $ f
            when (optVerbose opt) $ do
                comment opt $ "simplified formula:" `nl` indent 4 (Fm.pretty f')
            -- Having parseProblemFile return a list is a hack
            -- to allow for multiple TRSs in the same file
            -- for the commutation format in CoCo.
            ps <- concat <$> mapM P.parseProblemFile fs
            let sig = S.nubOrd $ ps >>= P.signature
                trss = map P.rules ps
            when (optVerbose opt) $ do
                comment opt $ "signature:" `nl` indent 4 (pretty sig)
            unless (all isLVTRS trss) $ do
                printResult "MAYBE"
                comment opt $ "input TRSs must be LV-TRSs"
                exitWith $ ExitFailure 1
            eval opt sig trss f f'
    main' _ _ = do
        hPutStrLn stderr "Error: expected a formula and at least one TRS\n"
        usage (ExitFailure 1)

comment :: Options -> Doc ann -> IO ()
comment opt = Text.putStr . tweak .
    renderLazy . layoutPretty defaultLayoutOptions
  where
    tweak = (<> "\n")
--         | certOn opt = Text.unlines . map ("; " <>) . Text.lines
--         | otherwise = (<> "\n")

printResult = Text.putStr . (<> "\n") . renderLazy
        . layoutPretty defaultLayoutOptions


-- | We strip outermost negations and existentials so that we can
--
--  * avoid some unnecessary projections; and
--  * produce witnesses
eval :: Options -> Signature String String -> [[R.Rule String String]] -> Fm.Formula -> Fm.Formula -> IO ()
eval opt sig trss f0 f1 = go False 0 f1
  where
    go :: Bool -> Int -> Fm.Formula -> IO ()
    go b 0 (Fm.FNot f) = go (not b) 0 f
    go b n (Fm.FExists f) = go b (n + 1) f
    -- If top-level is formula for difference, use faster subset check.
    go b _ f@(Fm.FAnd [fl, Fm.FNot fr])= do
        let ((i, (rrnl, rrnr), vs), ss) = Fm.evalM $ do
                r1 <- Fm.mkFormula (optInfo opt) sig extSig trss fl
                r2 <- Fm.mkFormula (optInfo opt) sig extSig trss fr
                let ([rrnl, rrnr], vs) = Fm.fit [r1,r2]
                -- Add inferences leading up to f1 (which has f as subformula).
                -- Note that thanks to laziness, the underlying tree automaton
                -- is never computed.
                (j, _, _) <- Fm.mkFormula False sig extSig trss f1
                -- log simplification step
                k <- Fm.nextI
                Fm.tellM (k, Fm.INNFPlus j, f0, [])
                return (k, (rrnl, rrnr), vs)
            tal = RRn.ta rrnl
            tar = RRn.ta rrnr
        if not (optWitness opt)
        then do
            -- no witness requested -> just check subset
            let e = TA.subset tal tar
            result $ if b then e else not e
            certificate ss (if b then not e else e) i
        else -- witness generation enabled -> ask for non-subset witness
            case TA.witnessDiff tal tar (length . catMaybes) of
                Nothing -> do
                     result b
                     certificate ss (not b) i
                Just w -> do
                     result (not b)
                     comment opt $ nl "formula body / witness:" $
                         indent 4 $ nl (Fm.pretty f) $ vcat $ do
                             (i, v) <- zip [0..] (S.toList vs)
                             return $ pretty v <+> "=" <+> (hang 0 $
                                 pretty (fromJust (decode i w)))
                     certificate ss b i
    go b _ f = do
        let ((i, rrn, vs), ss) = Fm.evalM $ do
                (_, rrn, vs) <- Fm.mkFormula (optInfo opt) sig extSig trss f
                -- Add inferences leading up to f1 (which has f as subformula).
                -- Note that thanks to laziness, the underlying tree automaton
                -- is never computed.
                (j, _, _) <- Fm.mkFormula False sig extSig trss f1
                -- log simplification step
                k <- Fm.nextI
                Fm.tellM (k, Fm.INNFPlus j, f0, [])
                return (k, rrn, vs)
            ta = RRn.ta rrn
        if not (optWitness opt)
        then do
            -- no witness requested -> just check emptiness
            let e = TA.isEmpty ta
            result $ if b then e else not e
            certificate ss (if b then not e else e) i
        else do
            -- witness generation enabled -> ask for non-emptiness witness
            case TA.witness ta (length . catMaybes) of
                Nothing -> do
                     result b
                     certificate ss (not b) i
                Just w -> do
                     result (not b)
                     comment opt $ nl "formula body / witness:" $
                         indent 4 $ nl (Fm.pretty f) $ vcat $ do
                             (i, v) <- zip [0..] (S.toList vs)
                             return $ pretty v <+> "=" <+> (hang 0 $
                                 pretty (fromJust (decode i w)))
                     certificate ss b i

    extSig = [[ (c, [], s)
              | i <- [0..]
              , let c = '_':(s ++ show i)
              , (c, [], s) `notElem` sig
              ] | s <- sorts sig]

    result b = printResult (if b then "YES" else "NO")
    -- Previously:
    --   result b = comment opt $ if b then "YES" else "NO"
    -- but for CoCo first line has to be the result without leading ';'

    certificate ss e i =
        case optCertificate opt of
          Nothing -> return ()
          Just h -> do
            hPrint h $ Fm.output $
                Fm.Certificate ss (if e then Fm.Empty else Fm.Nonempty) i
            hFlush h

-- | Decode a component of an /RRn/ encoded term.
decode :: Int -> T.Term [Maybe f] v -> Maybe (T.Term f v)
decode i = go where
    go (T.Fun fs ts) = case fs !! i of
         Nothing -> Nothing
         Just f -> Just (T.Fun f (catMaybes (map go ts)))
    go T.Var{} = error "decode: expected ground term"

-- pretty-print terms
instance (Pretty f, Pretty v) => Pretty (T.Term f v) where
    pretty (T.Var x) = pretty x
    pretty (T.Fun f ts) = pretty f <> encloseSep' "(" ")" "," (map pretty ts)
