-- | Basic formula manipulation.
module Formula.Basics (
    fvars,
    arity,
    rename,
    unrestrict,
    nnf,
    nnfPlus,
) where

import Formula.Type
import qualified Util.Set as S

-- | Compute free variables of a formula.
fvars :: Formula -> S.Set Int
fvars f = S.fromList (go 0 f [])
  where
    var b x = if x >= b then (x - b :) else id
    go b (FRR1 _ x) = var b x
    go b (FRR2 _ x y) = var b x . var b y
    go b (FAnd fs) = foldr ((.) . go b) id fs
    go b (FOr fs) = foldr ((.) . go b) id fs
    go b (FNot f) = go b f
    go b (FExists f) = go (b + 1) f
    go b (FForall f) = go (b + 1) f
    -- derived
    go _ FTrue = id
    go _ FFalse = id
    go b (FExtendSig f _) = go b f
    go b (FRestrict f _) = go b f

-- | Compute arity of a formula.
--
-- > arity f == maximum (0 : map (+1) (S.toList (fvars f)))
arity :: Formula -> Int
arity (FRR1 _ x) = x + 1
arity (FRR2 _ x y) = max x y + 1
arity (FAnd fs) = maximum (0 : map arity fs)
arity (FOr fs) = maximum (0 : map arity fs)
arity (FNot f) = arity f
arity (FExists f) = max 0 (arity f - 1)
arity (FForall f) = max 0 (arity f - 1)
arity FTrue = 0
arity FFalse = 0
arity (FExtendSig f _) = arity f
arity (FRestrict f _) = arity f

-- | Rename free variables of a formula.
rename :: (Int -> Int) -> Formula -> Formula
rename vs f = go 0 f
  where
    var b x = if x >= b then b + vs (x - b) else x
    go b (FRR1 s x) = FRR1 s (var b x)
    go b (FRR2 r x y) = FRR2 r (var b x) (var b y)
    go b (FAnd fs) = FAnd (map (go b) fs)
    go b (FOr fs) = FOr (map (go b) fs)
    go b (FNot f) = FNot (go b f)
    go b (FExists f) = FExists (go (b + 1) f)
    go b (FForall f) = FForall (go (b + 1) f)
    -- derived
    go _ FTrue = FTrue
    go _ FFalse = FFalse
    go b (FExtendSig f n) = FExtendSig (go b f) n
    go b (FRestrict f ts) = FRestrict (go b f) ts

-- | Expand `FRestrict` constructs in formula.
unrestrict :: Formula -> Formula
unrestrict = go [0..]
  where
    -- so much boilerplate... maybe we should use generics?
    go is (FRR1 s x) = FRR1 (go1 is s) x
    go is (FRR2 r x y) = FRR2 (go2 is r) x y
    go is (FAnd fs) = FAnd (map (go is) fs)
    go is (FOr fs) = FOr (map (go is) fs)
    go is (FNot f) = FNot (go is f)
    go is (FExists f) = FExists (go is f)
    go is (FForall f) = FForall (go is f)
    -- derived
    go _  FTrue = FTrue
    go _  FFalse = FFalse
    go is (FExtendSig f n) = FExtendSig (go is f) n
    go is (FRestrict f ts) = go (map (is !!) ts ++ drop (length ts) is) f

    got is (Fwd t) = Fwd (is !! t)
    got is (Bwd t) = Bwd (is !! t)

    gog is (ARoot ts) = ARoot (map (got is) ts)
    gog is (GInv g) = GInv (gog is g)
    gog is (AUnion g1 g2) = AUnion (gog is g1) (gog is g2)
    gog is (ATrancl g) = ATrancl (gog is g)
    gog is (GTrancl g) = GTrancl (gog is g)
    gog is (AComp g1 g2) = AComp (gog is g1) (gog is g2)
    gog is (GComp g1 g2) = GComp (gog is g1) (gog is g2)
    -- derived
    gog is (GSteps ts) = GSteps (map (got is) ts)

    go1 _  R1Terms = R1Terms
    go1 is (R1NF ts) = R1NF (map (got is) ts)
    go1 is (R1Inf r) = R1Inf (go2 is r)
    go1 is (R1Proj i r) = R1Proj i (go2 is r)
    go1 is (R1Union s1 s2) = R1Union (go1 is s1) (go1 is s2)
    go1 is (R1Inter s1 s2) = R1Inter (go1 is s1) (go1 is s2)
    go1 is (R1Diff s1 s2) = R1Diff (go1 is s1) (go1 is s2)
    -- derived
    go1 is (R1Fin r) = R1Fin (go2 is r)

    go2 is (R2GTT g w x) = R2GTT (gog is g) w x
    go2 is (R2Diag s) = R2Diag (go1 is s)
    go2 is (R2Inv r) = R2Inv (go2 is r)
    go2 is (R2Prod s1 s2) = R2Prod (go1 is s1) (go1 is s2)
    go2 is (R2Union r1 r2) = R2Union (go2 is r1) (go2 is r2)
    go2 is (R2Inter r1 r2) = R2Inter (go2 is r1) (go2 is r2)
    go2 is (R2Diff r1 r2) = R2Diff (go2 is r1) (go2 is r2)
    go2 is (R2Comp r1 r2) = R2Comp (go2 is r1) (go2 is r2)
    -- derived
    go2 _  R2Eq = R2Eq
    go2 is (R2Reflc r) = R2Reflc (go2 is r)
    go2 is (R2Step ts) = R2Step (map (got is) ts)
    go2 is (R2StepEq ts) = R2StepEq (map (got is) ts)
    go2 is (R2ParStep ts) = R2ParStep (map (got is) ts)
    go2 is (R2Steps ts) = R2Steps (map (got is) ts)
    go2 is (R2StepsEq ts) = R2StepsEq (map (got is) ts)
    go2 is (R2StepsNF ts) = R2StepsNF (map (got is) ts)
    go2 is (R2RootStep ts) = R2RootStep (map (got is) ts)
    go2 is (R2RootStepEq ts) = R2RootStepEq (map (got is) ts)
    go2 is (R2RootSteps ts) = R2RootSteps (map (got is) ts)
    go2 is (R2RootStepsEq ts) = R2RootStepsEq (map (got is) ts)
    go2 is (R2NonRootStep ts) = R2NonRootStep (map (got is) ts)
    go2 is (R2NonRootStepEq ts) = R2NonRootStepEq (map (got is) ts)
    go2 is (R2NonRootSteps ts) = R2NonRootSteps (map (got is) ts)
    go2 is (R2NonRootStepsEq ts) = R2NonRootStepsEq (map (got is) ts)
    go2 is (R2Meet ts) = R2Meet (map (got is) ts)
    go2 is (R2Join ts) = R2Join (map (got is) ts)

-- | Convert formula to negation normal form.
nnf :: Formula -> Formula
nnf = nnfAux True

-- | Convert formula to negation normal form, plus trivial
-- simplification (unit, idempotency, zero for conjunction and disjunction)
nnfPlus :: Formula -> Formula
nnfPlus = nnfAux False

nnfAux :: Bool -> Formula -> Formula
nnfAux basic = go False where
    go b f@FRR1{} = if b then FNot f else f
    go b f@FRR2{} = if b then FNot f else f
    go b (FAnd fs) = (if b then fOr else fAnd) (map (go b) fs)
    go b (FOr fs) = (if b then fAnd else fOr) (map (go b) fs)
    go b (FNot f) = go (not b) f
    go b (FExists f) = (if b then FForall else FExists) (go b f)
    go b (FForall f) = (if b then FExists else FForall) (go b f)
    go b FTrue = if b then FFalse else FTrue
    go b FFalse = if b then FTrue else FFalse
    go b (FExtendSig f n) = FExtendSig (go b f) n
    go b (FRestrict f ts) = FRestrict (go b f) ts

    fOr = if basic then FOr else fin . S.nubOrd . (>>= flat) where
        fin [] = FFalse
        fin [f] = f
        fin fs = if FTrue `elem` fs then FTrue else FOr fs
        flat FFalse = []
        flat (FOr fs) = fs
        flat f = [f]

    fAnd = if basic then FAnd else fin . S.nubOrd . (>>= flat) where
        fin [] = FTrue
        fin [f] = f
        fin fs = if FFalse `elem` fs then FFalse else FAnd fs
        flat FTrue = []
        flat (FAnd fs) = fs
        flat f = [f]
