{-# LANGUAGE FlexibleContexts #-}

-- | FORT compatible formula parser
module Formula.FORT (
    fortFormula,
) where

import Formula.Type

import Text.Parsec
import Util.Parsec
-- import Text.Parsec.Char
import Control.Monad
import Data.List

-- | Parser for FORT formula.
fortFormula :: Parsec String () Formula
fortFormula = spaces *> formula [] <* eof

toks :: [String] -> Parsec String () ()
toks ts = void $ msum [try (string t) <* spaces | t <- ts]

name :: Parsec String () String
name = (:) <$> letter <*> many alphaNum <* spaces

names :: Parsec String () [String]
names = (:) <$> name <*> many (toks [","] *> name)

nat :: Parsec String () Int
nat = read <$> many1 digit <* spaces

-- <var>
var :: [String] -> Parsec String () Int
var vs = do
    p <- getPosition
    v <- name
    case elemIndex v vs of
        Just i -> return i
        Nothing -> sysUnexpectedAt p $
            "variable `" ++ v ++ "': closed formula expected"

-- <formula>
formula :: [String] -> Parsec String () Formula
formula vs = iff
  where
    fAnd [x] = x; fAnd xs = FAnd xs
    fOr [x] = x; fOr xs = FOr xs
    iff = foldr1 fIff <$> ((:) <$> imp <*> many (toks ["<=>", "⇔"] *> imp))
    imp = foldr1 fImplies <$> ((:) <$> dis <*> many (toks ["=>", "⇒"] *> dis))
    dis = fOr <$> ((:) <$> con <*> many (toks ["|", "∨"] *> con))
    con = fAnd <$> ((:) <$> neg <*> many (toks ["&", "∧"] *> neg))
    neg = FNot <$> (toks ["~", "¬"] *> neg) <|> bas
    bas = toks ["("] *> iff <* toks [")"] <|>
        property vs <|>
        (do q <- quantifier
            ws <- names
            toks ["("]
            f <- formula (reverse ws ++ vs)
            toks [")"]
            return $ iterate q f !! length ws) <|>
        (\ts f -> FRestrict f ts) <$>
             (toks ["["] *> ((:) <$> nat <*> many (toks [","] *> nat)) <*
             toks ["]"]) <*> bas <|>
        (\x r y -> FRR2 r x y) <$> var vs <*> relation <*> var vs

-- <relation>
relation :: Parsec String () RR2_rel
relation = (<?> "relation") $
    R2RootSteps [Fwd 0] <$ toks ["->e+", "->ε+", "→e+", "→ε+"] <|>
    R2RootStepsEq [Fwd 0] <$ toks ["->e*", "->ε*", "→e*", "→ε*"] <|>
    R2RootStepEq [Fwd 0] <$ toks ["->e=", "->ε=", "→e=", "→ε="] <|>
    R2RootStep [Fwd 0] <$ toks ["->e", "->ε", "→e", "→ε"] <|>
    R2NonRootSteps [Fwd 0] <$ toks ["->be+", "->⌄ε+", "→be+", "→⌄ε+"] <|>
    R2NonRootStepsEq [Fwd 0] <$ toks ["->be*", "->⌄ε*", "→be*", "→⌄ε*"] <|>
    R2NonRootStepEq [Fwd 0] <$ toks ["->be=", "->⌄ε=", "→be=", "→⌄ε="] <|>
    R2NonRootStep [Fwd 0] <$ toks ["->be", "->⌄ε", "→be", "→⌄ε"] <|>
    R2Steps [Fwd 0] <$ toks ["->+", "→+"] <|>
    R2StepsEq [Fwd 0] <$ toks ["->*", "→*"] <|>
    R2StepEq [Fwd 0] <$ toks ["->=", "→="] <|>
    R2StepsNF [Fwd 0] <$ toks ["->!", "→!"] <|>
    R2Step [Fwd 0] <$ toks ["->", "→"] <|>
    R2ParStep [Fwd 0] <$ toks ["-||->", "⇻"] <|>

    R2StepsEq [Fwd 0, Bwd 0] <$ toks ["<->*", "↔*"] <|>
    R2Step [Fwd 0, Bwd 0] <$ toks ["<->", "↔"] <|>

    R2RootSteps [Bwd 0] <$ toks ["+e<-", "+ε<-", "+e←", "+ε←"] <|>
    R2RootStepsEq [Bwd 0] <$ toks ["*e<-", "*ε<-", "*e←", "*ε←"] <|>
    R2RootStepEq [Bwd 0] <$ toks ["=e<-", "=ε<-", "=e←", "=ε←"] <|>
    R2RootStep [Bwd 0] <$ toks ["e<-", "ε<-", "e←", "ε←"] <|>
    R2NonRootSteps [Bwd 0] <$ toks ["+be<-", "+⌄ε<-", "+be←", "+⌄ε←"] <|>
    R2NonRootStepsEq [Bwd 0] <$ toks ["*be<-", "*⌄ε<-", "*be←", "*⌄ε←"] <|>
    R2NonRootStepEq [Bwd 0] <$ toks ["=be<-", "=⌄ε<-", "=be←", "=⌄ε←"] <|>
    R2NonRootStep [Bwd 0] <$ toks ["be<-", "⌄ε<-", "be←", "⌄ε←"] <|>
    R2Steps [Bwd 0] <$ toks ["+<-", "+←"] <|>
    R2StepsEq [Bwd 0] <$ toks ["*<-", "*←"] <|>
    R2StepEq [Bwd 0] <$ toks ["=<-", "=←"] <|>
    R2StepsNF [Bwd 0] <$ toks ["!<-", "!←", "<-!", "←!"] <|>
    R2ParStep [Bwd 0] <$ toks ["<-||-", "⇺"] <|>
    R2Step [Bwd 0] <$ toks ["<-", "←"] <|>

    R2Join [Fwd 0] <$ toks ["join", "↓"] <|>
    R2Meet [Fwd 0] <$ toks ["meet", "↑"] <|>
    R2Eq <$ toks ["="]

-- <quantifier>
quantifier :: Parsec String () (Formula -> Formula)
quantifier =
    FForall <$ toks ["forall", "∀"] <|>
    FExists <$ toks ["exists", "∃"]

-- <property>
property :: [String] -> Parsec String () Formula
property vs =
    try (property1 vs) <|>
    gunc <$ toks ["GUNC"] <|>
    unc <$ toks ["UNC"] <|>
    gcr <$ toks ["GCR"] <|>
    cr <$ toks ["CR"] <|>
    gnfp <$ toks ["GNFP"] <|>
    nfp <$ toks ["NFP"] <|>
    wn <$ toks ["WN"] <|>
    sn <$ toks ["SN"] <|>
    (gcom <$ toks ["GCom"] <|>
     com <$ toks ["Com"] <|>
     glcom <$ toks ["GLCom"] <|>
     lcom <$ toks ["LCom"]) <*>
        (toks ["("] *> ixs) <*> (toks [","] *> ixs <* toks [")"]) <|>
    (gun <$ toks ["GUN"] <|> -- why does this take a relation?
     un <$ toks ["UN"] <|> -- why does this take a relation?
     gwcr <$ toks ["GWCR"] <|>
     wcr <$ toks ["WCR"] <|>
     gscr <$ toks ["GSCR"] <|>
     scr <$ toks ["SCR"]) <*>
        (toks ["("] *> gtt_rel <* toks [")"] <|> return (R2Step [Fwd 0], R2StepsEq [Fwd 0])) <|>
    (gdiamond <$ toks ["Gdiamond"] <|>
     diamond <$ toks ["diamond"]) <*>
        (toks ["("] *> formula2 <* toks [")"] <|> return (R2Step [Fwd 0]))
  where
    ixs = toks ["["] *> ((:) <$> nat <*> (many (toks [","] *> nat))) <* toks ["]"]
    glcom' (r, rs) (s, ss) = FForall (FForall (fImplies (FRR2 (R2Comp (R2Inv s) r) 0 1) (FRR2 (R2Comp rs (R2Inv ss)) 0 1)))
    gcom' rs ss = FForall (FForall (fImplies (FRR2 (R2Comp (R2Inv ss) rs) 0 1) (FRR2 (R2Comp rs (R2Inv ss)) 0 1)))
    gcr = FForall (FForall (
        fImplies (FRR2 (R2Comp (R2Step [Bwd 0]) (R2StepsEq [Fwd 0])) 0 1)
                 (FRR2 (R2Comp (R2StepsEq [Fwd 0]) (R2StepsEq [Bwd 0])) 0 1)))
    cr = FExtendSig gcr 2
    gwcr (r, rs) = glcom' (r, rs) (r, rs)
    wcr (r, rs) = FExtendSig (gwcr (r, rs)) 2
    gscr (r, rs) = FForall (FForall (fImplies (FRR2 (R2Comp (R2Inv r) r) 0 1) (FRR2 (R2Comp (R2Reflc r) (R2Inv rs)) 0 1)))
    scr (r, rs) = FExtendSig (gscr (r, rs)) 2
    gun _ = let rs = R2StepsNF [Fwd 0] in FForall (FForall (fImplies (FRR2 (R2Comp (R2Inv rs) rs) 0 1) (FRR2 R2Eq 0 1)))
    un a = FExtendSig (gun a) 2
    gunc = FForall (FForall (
        fImplies (FAnd [FRR2 (R2Prod (R1NF [Fwd 0]) (R1NF [Fwd 0])) 0 1, FNot (FRR2 R2Eq 0 1)])
                 (FNot (FRR2 (R2StepsEq [Fwd 0, Bwd 0]) 0 1))))
    unc = FExtendSig gunc 2
    gnfp = FForall (FForall (
        fImplies (FAnd [FRR2 (R2Prod R1Terms (R1NF [Fwd 0])) 0 1, FNot (FRR2 (R2StepsEq [Fwd 0]) 0 1)])
                 (FNot (FRR2 (R2StepsEq [Fwd 0, Bwd 0]) 0 1))))
    nfp = FExtendSig gnfp 2
    gdiamond r =  FForall (FForall (fImplies (FRR2 (R2Comp (R2Inv r) r) 0 1) (FRR2 (R2Comp r (R2Inv r)) 0 1)))
    diamond r = FExtendSig (gdiamond r) 2
    --wn = FForall (FExists (FRR2 (R2StepsNF [Fwd 0]) 1 0))
    wn = FForall (FExists (FAnd [FRR2 (R2StepsEq [Fwd 0]) 1 0, FRR1 (R1NF [Fwd 0]) 0]))
    sn = FAnd [FForall (FNot (FRR2 (R2Steps [Fwd 0]) 0 0)), FForall (FRR1 (R1Fin (R2StepsEq [Fwd 0])) 0)]
    glcom is js = glcom' (R2Step (map Fwd is), R2StepsEq (map Fwd is)) (R2Step (map Fwd js), R2StepsEq (map Fwd js))
    lcom is js = FExtendSig (glcom is js) 2
    gcom is js = gcom' (R2StepsEq (map Fwd is)) (R2StepsEq (map Fwd js))
    com is js = FExtendSig (gcom is js) 2

property1 :: [String] -> Parsec String () Formula
property1 vs =
    (cr <$ toks ["CR"] <|>
     wcr <$ toks ["WCR"] <|>
     wn <$ toks ["WN"] <|>
     nfp <$ toks ["NFP"] <|>
     sn <$ toks ["SN"] <|>
     nf <$ toks ["NF"]) <*> (toks ["("] *> var vs <* toks [")"]) <|>
    (scr <$ toks ["SCR"] <|>
     un <$ toks ["UN"]) <*>
        (toks ["("] *> (gtt_rel <* toks [","] <|> return (R2Step [Fwd 0], R2StepsEq [Fwd 0]))) <*>
        (var vs <* toks [")"]) <|>
    (diamond <$ toks ["diamond"] <|>
     inf <$ toks ["Inf"] <|>
     fin <$ toks ["Fin"]) <*>
        (toks ["("] *> (formula2 <* toks [","] <|> return (R2Step [Fwd 0]))) <*>
        (var vs <* toks [")"])
  where
    r = R2Step [Fwd 0]
    rs = R2StepsEq [Fwd 0]
    cr v = FForall (FForall (fImplies (FAnd [FRR2 rs (v+2) 0, FRR2 rs (v+2) 1]) (FRR2 (R2Comp rs (R2Inv rs)) 0 1)))
    wcr v = FForall (FForall (fImplies (FAnd [FRR2 r (v+2) 0, FRR2 r (v+2) 1]) (FRR2 (R2Comp rs (R2Inv rs)) 0 1)))
    wn v = FExists (FRR2 (R2StepsNF [Fwd 0]) (v+1) 0)
    nfp v = FForall (fImplies (FRR2 (R2Inter (R2StepsEq [Fwd 0, Bwd 0]) (R2Prod R1Terms (R1NF [Fwd 0]))) (v+1) 0) (FRR2 (R2StepsEq [Fwd 0]) (v+1) 0))
    sn v = FAnd [FForall (fImplies (FRR2 (R2StepsEq [Fwd 0]) (v+1) 0) (FNot (FRR2 (R2Steps [Fwd 0]) 0 0))), FRR1 (R1Fin (R2StepsEq [Fwd 0])) v]
    nf v = FRR1 (R1NF [Fwd 0]) v
    scr (r, rs) v = FForall (FForall (fImplies (FAnd [FRR2 r (v+2) 0, FRR2 r (v+2) 1]) (FRR2 (R2Comp (R2Reflc r) (R2Inv rs)) 0 1)))
    un _ v = let rs = R2StepsNF [Fwd 0] in FForall (FForall (fImplies (FAnd [FRR2 rs (v+2) 0, FRR2 rs (v+2) 1]) (FRR2 R2Eq 0 1)))
    diamond r v = FForall (FForall (fImplies (FAnd [FRR2 r (v+2) 0, FRR2 r (v+2) 1]) (FRR2 (R2Comp r (R2Inv r)) 0 1)))
    fin r v = FRR1 (R1Fin r) v
    inf r v = FRR1 (R1Inf r) v

-- <formula2>
formula2 :: Parsec String () RR2_rel
formula2 = iff
  where
    iff = foldr1 r2Iff <$> ((:) <$> imp <*> many (toks ["<=>", "⇔"] *> imp))
    imp = foldr1 r2Implies <$> ((:) <$> dis <*> many (toks ["=>", "⇒"] *> dis))
    dis = foldr1 R2Union <$> ((:) <$> con <*> many (toks ["|", "∨"] *> con))
    con = foldr1 R2Inter <$> ((:) <$> neg <*> many (toks ["&", "∧"] *> neg))
    neg = r2Neg <$> (toks ["~", "¬"] *> neg) <|> bas
    bas = toks ["("] *> iff <* toks [")"] <|> relation

-- <gtt_rel>
gtt_rel :: Parsec String () (RR2_rel, RR2_rel)
gtt_rel =
    (R2Step [Fwd 0], R2StepsEq [Fwd 0]) <$ toks ["->", "→"] <|>
    (R2ParStep [Fwd 0], R2StepsEq [Fwd 0]) <$ toks ["-||->", "⇻"]
