-- | GTT composition and transitive closure.
module GTT.Compose (
    gcomp,
    acomp,
    gtrancl,
    atrancl,
) where

import Prelude as P
import qualified TA
import GTT.Type as GTT
import Util.Saturate
import qualified Data.Set as S

-- | GTT composition.
--
-- May destroy root-cleanliness.
gcomp :: (Eq f, Ord p, Ord q) => GTT f p -> GTT f q -> GTT f (Either p q)
gcomp g1 g2 = g' `GTT.merge` GTT {
    leftTA  = TA.empty{ TA.eps = epsl },
    rightTA = TA.empty{ TA.eps = P.map (\(p, q) -> (q, p)) epsl }
} where
    (g1', g2') = (GTT.map id Left g1, GTT.map id Right g2)
    g' = GTT.merge g1' g2'
    epsl = S.toList (saturate sat0 sat1)
    sat0 = gc_sat0 g1' g2'
    sat1 = gc_sat1 g1' g2'

-- | Anchored GTT composition.
--
-- Assumes and preserves root-cleanliness (see `anchor`).
acomp :: (Eq f, Ord p, Ord q) => GTT f p -> GTT f q -> GTT f (Either p q)
acomp g1 g2 = g' `GTT.merge` GTT {
    leftTA  = TA.empty{ TA.eps = epsl },
    rightTA = TA.empty{ TA.eps = P.map (\(p, q) -> (q, p)) epsl }
} where
    (g1', g2') = (GTT.map id Left g1, GTT.map id Right g2)
    g' = GTT.merge g1' g2'
    ifs1 = (`S.member` interface g1')
    ifs2 = (`S.member` interface g2')
    epsl = P.filter (\(p, q) -> ifs1 p && ifs2 q) (S.toList (saturate sat0 sat1))
    sat0 = gc_sat0 g1' g2'
    sat1 = gc_sat1 g1' g2'

-- | GTT transitive closure.
--
-- May destroy root-cleanliness.
gtrancl :: (Eq f, Ord q) => GTT f q -> GTT f q
gtrancl g = g `GTT.merge` GTT {
    leftTA  = TA.empty{ TA.eps = epsl },
    rightTA = TA.empty{ TA.eps = P.map (\(p, q) -> (q, p)) epsl }
} where
    epsl = S.toList (saturate sat0 sat1)
    sat0 = gc_sat0 g g
    sat1 e@(p, q) es =
        [(p, q'') | (q', q'') <- S.toList es, q == q'] ++
        [(p'', q) | (p'', p') <- S.toList es, p' == p] ++
        gc_sat1 g g e es

-- | Anchored GTT transitive closure.
--
-- Requires and preserves root-cleanliness.
atrancl :: (Eq f, Ord q) => GTT f q -> GTT f q
atrancl g = g `GTT.merge` GTT {
    leftTA  = TA.empty{ TA.eps = epsl },
    rightTA = TA.empty{ TA.eps = P.map (\(p, q) -> (q, p)) epsl }
} where
    ifs = (`S.member` interface g)
    epsl = P.filter (\(p, q) -> ifs p && ifs q) (S.toList (saturate sat0 sat1))
    sat0 = gc_sat0 g g
    sat1 e@(p, q) es =
        [(p, q'') | (q', q'') <- S.toList es, q == q'] ++
        [(p'', q) | (p'', p') <- S.toList es, p' == p] ++
        gc_sat1 g g e es

-- Auxiliary functions shared between composition and transitive closure.
gc_sat0 :: (Eq f, Ord q) => GTT f q -> GTT f q -> [(q, q)]
gc_sat0 GTT{ rightTA = a1 } GTT{ leftTA = a2 } =
    [(p, q) | (f, [], p) <- TA.trans a1, (f', [], q) <- TA.trans a2, f == f']

gc_sat1 :: (Eq f, Ord q) =>
    GTT f q -> GTT f q -> (q, q) -> S.Set (q, q) -> [(q, q)]
gc_sat1 GTT{ rightTA = a1 } GTT{ leftTA = a2 } (p, q) es =
    [(p'', q) | (p', p'') <- TA.eps a1, p == p'] ++
    [(p, q'') | (q', q'') <- TA.eps a2, q == q'] ++
    [(p', q')
    | (f, ps, p') <- TA.trans a1,
      (f', qs, q') <- TA.trans a2,
      (f, length ps) == (f', length qs),
      (p, q) `elem` zip ps qs,
      all (`S.member` es) (zip ps qs)]
