{-# LANGUAGE OverloadedStrings #-}

module Main (
    main,
) where

import qualified Formula as Fm
import qualified Formula.Automata as Fm
import qualified Formula.Simplify as Fm
import qualified Formula.FORT as Fm
import qualified RRn
import qualified TA
import Util.Pretty
import Util.Rewriting
import Util.Signature

import qualified Problem as P
import qualified Data.Rewriting.Rule as R
import qualified Data.Rewriting.Term as T
import qualified Data.Rewriting.Term.Parse as TP
import qualified ArgumentParser as AP

import Satisfiability
import System.IO
import System.Exit
import System.Environment
import Control.Monad
import Data.List

import Data.Text.Prettyprint.Doc hiding (list)
import Data.Text.Prettyprint.Doc.Render.Text
import Text.Parsec (parse, eof)

import System.Console.GetOpt


data Options = Options {
    optSig :: Bool,
    optDepth :: Int,
    optRules :: Int,
    optVars :: Int
}

opt0 = Options {
    optSig = False,
    optDepth = 3,
    optRules = 3,
    optVars = 1
}

options :: [OptDescr (Options -> IO Options)]
options = [
    Option "h?"  ["help"]
        (NoArg (\_ -> usage ExitSuccess))
        "print usage information",
    Option "a" ["arity"]
        (NoArg (\o -> return $ o{ optSig = True }))
        "signature with arities",
    Option "d" ["depth"]
        (ReqArg (\d o -> return $ o{ optDepth = 
            case parse (AP.nat <* eof) "depth" d of
                Left e -> error (show e)
                Right f -> f }) "NUM")
        ("maximum depth, default: " ++ show (optDepth opt0)),
    Option "r" ["rules"]
        (ReqArg (\r o -> return $ o{ optRules = 
            case parse (AP.nat <* eof) "numRules" r of
                Left e -> error (show e)
                Right f -> f }) "NUM")
        ("maximum number of rules, default: " ++ show (optRules opt0)),
    Option "v" ["vars"]
        (ReqArg (\v o -> return $ o{ optVars = 
            case parse (AP.nat <* eof) "numVars" v of
                Left e -> error (show e)
                Right f -> f }) "NUM")
        ("maximum number of variables, default: " ++ show (optVars opt0))  
    ]


usage :: ExitCode -> IO a
usage e = do
    hPutStrLn stderr $
        usageInfo header options
    exitWith e
  where
    header = unlines [
        "Usage: synth [OPTION...] SIGNATURE FORMULA"
     ]

main :: IO()
main = do  
    args <- getArgs
    case getOpt Permute options args of
        (_,_ , e:_) -> do
            hPutStrLn stderr $ "Error: " ++ e
            usage (ExitFailure 1)
        (os, fs, []) -> do
            opt <- foldl (>>=) (return opt0) os  
            main' opt fs    
    where 
        main' opt [] = do
            hPutStrLn stderr "Error: expected a signature and a formula\n"
            usage (ExitFailure 1)
        main' opt fs = do
            let [sig,formula] = fs
            let sig' = getSig opt sig
            let depth' = optDepth opt
            let numRules' = optRules opt
            let numVars' = optVars opt
            case parse Fm.fortFormula "formula" formula of
                Left e -> do 
                    hPutStrLn stderr $ "Error: formula could not be parsed"
                    return()
                Right f -> do
                    searchTrs sig' f depth' numVars' numRules'
        

getSig :: Options -> String -> [(String,Int)]
getSig opt sig = if not (optSig opt) 
    then do
    case parse (AP.signatureParser <* eof) "sig" sig of
        Left e -> error (show e)
        Right f -> f
    else do
    case parse (AP.intListParser <* eof) "sig" sig of
        Left e -> error (show e)
        Right f ->
            generateSig f

generateSig :: [Int] -> [(String,Int)] 
generateSig as = zip (map (\i -> "c" ++ show i) [0 .. (length as)]) as

searchTrs :: [(String, Int)] -> Fm.Formula -> Int -> Int -> Int -> IO()
searchTrs sig formula d v r = searchTrsUntilDepth sig formula (d+1) v r [] [] []

searchTrsUntilDepth :: [(String, Int)] -> Fm.Formula -> Int -> Int -> Int -> [T.Term String String] -> [T.Term String String] -> [R.Rule String String] -> IO()
searchTrsUntilDepth _ _ 0 _ _ _ _ _ =  putStrLn ("\nNo TRS found within maximum depth.\n")
searchTrsUntilDepth sig formula d v r allTerms newTerms rules =    
    let newTerms' = getNewTerms sig v newTerms allTerms
        allTerms' = allTerms ++ newTerms' --maybe nub
        newRules = getNewRules sig newTerms' allTerms v
        newTrss = getNewTrss rules newRules r
        allRules = rules ++ newRules
        foundTrs = checkGiven sig formula newTrss 
    in
    if null foundTrs then do
        putStrLn("trs: " ++ show (Fm.pretty newTrss))
        searchTrsUntilDepth sig formula (d-1) v r allTerms' newTerms' allRules
    else do 
        putStrLn("trs: " ++ show (Fm.pretty newTrss))
        putStrLn("\n--- Found TRS --- \n" ++ show (Fm.pretty (head foundTrs)) ++ "\n") 

getNewTerms :: [(String, Int)] -> Int -> [T.Term String String] -> [T.Term String String] -> [T.Term String String]
getNewTerms sig numVars [] [] = makeConst sig ++ makeVar numVars
getNewTerms sig _ newTerms allTerms =
        let funL = getFuns sig in
        getNewFunTerms funL newTerms allTerms

getNewFunTerms :: [(String, Int)] -> [T.Term String String] -> [T.Term String String] -> [T.Term String String] 
getNewFunTerms [] _ _ = []
getNewFunTerms _ [] _ = []
getNewFunTerms (f:fs) newTerms allTerms =
    let permuts = getPermuts (snd f) newTerms allTerms
        newTerms' = createTerm (fst f) permuts 
    in
    newTerms' ++ getNewFunTerms fs newTerms allTerms

getPermuts :: Int -> [T.Term String String] -> [T.Term String String] -> [[T.Term String String]]
getPermuts 1 newTerms _ = replicateM 1 newTerms
getPermuts arity _ allTerms = replicateM arity allTerms

getNewTrss :: [R.Rule String String] -> [R.Rule String String] -> Int -> [[R.Rule String String]]
getNewTrss _ newRules 1 = toList newRules
getNewTrss oldRules newRules numRules = filter (\x -> length x <= numRules) (subsequences (oldRules ++ newRules)) 
-- TODO old rules are checked too often (ensure that at least one new rule is in subsequence)

getNewRules :: [(String, Int)] -> [T.Term String String] -> [T.Term String String] -> Int -> [R.Rule String String]
getNewRules sig newTerms oldTerms numVars = 
    let newLinearTerms = [i | i <- filter (\x -> T.isLinear x) newTerms]
        oldLinearTerms = [i | i <- filter (\x -> T.isLinear x) oldTerms]
        newRules = generateNewRules newLinearTerms oldLinearTerms 
    in  makeRules newRules

generateNewRules :: [T.Term String String] -> [T.Term String String] -> [(T.Term String String, T.Term String String)]
generateNewRules newTerms oldTerms =
    let allGround = filter (\x -> T.isGround x) (newTerms ++ oldTerms)
        newGround = filter (\x -> T.isGround x) newTerms 
        oldTerms' = filter (\x -> not (T.isVar x)) oldTerms
        newTerms' = filter (\x -> not (T.isVar x)) newTerms
    in
    [(x, y) | x <- newTerms', y <- allGround] ++ [(x, y) | x <- oldTerms', y <- newGround]

makeRules :: [(T.Term String String, T.Term String String)] -> [R.Rule String String]
makeRules trss = map (\trs -> R.Rule (fst trs) (snd trs)) trss

-- input: fun, permuts; output: terms
createTerm :: String -> [[T.Term String String]] -> [T.Term String String]
createTerm fun permuts = map (\i -> T.Fun fun i) permuts

makeVar :: Int -> [T.Term String String]
makeVar num = map (\x -> T.Var x) (map (\i -> "x" ++ show i) [0 .. (num-1)])

makeConst :: [(String, Int)]  -> [T.Term String String]
makeConst sig = map (\i -> T.Fun i []) (map fst (filter (\x -> snd x == 0) sig))

getFuns :: [(String, Int)] -> [(String, Int)]
getFuns sig = filter (\x -> not (snd x == 0)) sig

toList :: [a] -> [[a]]
toList as = map (\a -> [a]) as 

-- satisfies :: (Ord f, Ord v) => [(f, Int)] -> [[R.Rule f v]] -> Fm.Formula -> Bool
checkGiven :: [(String, Int)] -> Fm.Formula -> [[R.Rule String String]] -> [[R.Rule String String]]
checkGiven sig formula trss = filter (\trs -> satisfies sig [trs] formula) trss

-- pretty-print terms
instance (Pretty f, Pretty v) => Pretty (R.Rule f v) where
    pretty (R.Rule lhs rhs) = pretty lhs <> " -> " <> pretty rhs

-- pretty-print terms
instance (Pretty f, Pretty v) => Pretty (T.Term f v) where 
    pretty (T.Var x) = pretty x
    pretty (T.Fun f ts) = pretty f <> encloseSep' "(" ")" "," (map pretty ts)