module Main where

import System.Environment
import System.Exit
import Text.Parsec (parse)

import qualified Data.Rewriting.Term as T
import qualified Data.Rewriting.Rule as R

import Formula.FORT
import Formula.Parse
import Formula.Output
import Util.Set (nubOrd)
import qualified Formula.Type as Fm
import qualified Problem as P

import qualified Data.Char as Char

import FOR as FOR

usage :: IO ()
usage = do
  putStrLn "Usage: ./fortify ANSWER FORMULA CERTIFICAT TRS_FILES.."

getAnswer s = case map Char.toLower s of
  "yes" -> return True
  "no" -> return False
  _ -> do
    putStrLn "ANSWER must be either YES or NO"
    usage
    exitWith (ExitFailure 1)

main :: IO ()
main = do
  args <- getArgs
  case args of
    (answer:form:crtF:fs@(_:_)) -> do
      crt' <- readFile crtF
      a <- getAnswer answer
      let (f, crt) =
            case do
              f <- parse fortFormula "Formula" form
              crt <- parse certificate "Certificate" crt'
              return (f, crt)
            of
              Left e -> error (show e)
              Right r -> r
      ps <- concat <$> mapM P.parseProblemFile fs
      let b = check ps a f crt
      case b of
        Just True  -> putStrLn "Certified"
        Just False -> putStrLn "Wrong result"
        Nothing -> putStrLn "Invalid Certificate"
    _ -> usage

type Answer = Bool

check :: [P.Problem s String String] -> Answer -> Fm.Formula -> Fm.Certificate -> Maybe Bool
check ps answer f crt = check_certificate_string sig' trss' answer f' crt'
    where
        sig' = fset_of_list . map (\(f,i) -> (stringToInts f, nat_of_int i)) $ sig
        sig = nubOrd [(f,length ss) | p <- ps, (f, ss, s) <- P.signature p]

        trss' = [fset_of_list [ (term2FORTerm lhs', term2FORTerm rhs')
                              | R.Rule lhs rhs <- trs
                              , let lhs' = T.map stringToInts stringToInts lhs
                              , let rhs' = T.map stringToInts stringToInts rhs
                              ]
                | trs <- trss
                ]
        trss = [P.rules p | p <- ps]

        f' = form2FORform f
        crt' = crt2FORCrt crt

charToInt :: Char -> Integer
charToInt = fromIntegral . Char.ord

stringToInts :: String -> [Integer]
stringToInts = map charToInt

term2FORTerm :: T.Term f v -> FOR.Term f v
term2FORTerm (T.Var x) = FOR.Var x
term2FORTerm (T.Fun f ts) = FOR.Fun f (map term2FORTerm ts)

trs2FORTrs :: Fm.Trs -> FOR.Ftrs
trs2FORTrs (Fm.Fwd i) = FOR.Fwd (nat_of_int i)
trs2FORTrs (Fm.Bwd i) = FOR.Bwd (nat_of_int i)

pos2FORPos :: Fm.Pos_step -> FOR.Pos_step
pos2FORPos Fm.PRoot = FOR.PRoot
pos2FORPos Fm.PNonRoot = FOR.PNonRoot
pos2FORPos Fm.PAny = FOR.PAny

ext2FORExt :: Fm.Ext_step -> FOR.Ext_step
ext2FORExt e = case e of
    Fm.ESingle -> FOR.ESingle
    Fm.EParallel -> FOR.EParallel
    Fm.EStrictParallel -> FOR.EStrictParallel

gtt2FORGtt :: Fm.GTT_rel -> FOR.Gtt_rel FOR.Ftrs
gtt2FORGtt r = case r of
    Fm.ARoot trss -> FOR.ARoot (map trs2FORTrs trss)
    Fm.GInv gtt1 -> FOR.GInv (gtt2FORGtt gtt1)
    Fm.AUnion gtt1 gtt2 -> FOR.AUnion (gtt2FORGtt gtt1) (gtt2FORGtt gtt2)
    Fm.ATrancl gtt1 -> FOR.ATrancl (gtt2FORGtt gtt1)
    Fm.GTrancl gtt1 -> FOR.GTrancl (gtt2FORGtt gtt1)
    Fm.AComp gtt1 gtt2 -> FOR.AComp (gtt2FORGtt gtt1) (gtt2FORGtt gtt2)
    Fm.GComp gtt1 gtt2 -> FOR.GComp (gtt2FORGtt gtt1) (gtt2FORGtt gtt2)
    Fm.GSteps trss -> FOR.gSteps (map trs2FORTrs trss)

rr12FORRr1 :: Fm.RR1_rel -> FOR.Rr1_rel Ftrs
rr12FORRr1 r = case r of
    Fm.R1Terms -> FOR.R1Terms
    Fm.R1NF trss -> FOR.R1NF (map trs2FORTrs trss)
    Fm.R1Inf rr2 -> FOR.R1Inf (rr22FORRr2 rr2)
    Fm.R1Proj i rr2 -> FOR.R1Proj (nat_of_int i) (rr22FORRr2 rr2)
    Fm.R1Union rr11 rr12 -> FOR.R1Union (rr12FORRr1 rr11) (rr12FORRr1 rr12)
    Fm.R1Inter rr11 rr12 -> FOR.R1Inter (rr12FORRr1 rr11) (rr12FORRr1 rr12)
    Fm.R1Diff rr11 rr12 -> FOR.R1Diff (rr12FORRr1 rr11) (rr12FORRr1 rr12)
    Fm.R1Fin rr2 -> FOR.r1Fin (rr22FORRr2 rr2)

rr22FORRr2 :: Fm.RR2_rel -> FOR.Rr2_rel Ftrs
rr22FORRr2 r = case r of
    Fm.R2GTT ggt pos ext -> FOR.R2GTT_Rel (gtt2FORGtt ggt) (pos2FORPos pos) (ext2FORExt ext)
    Fm.R2Diag rr1 -> FOR.R2Diag (rr12FORRr1 rr1)
    Fm.R2Prod rr11 rr12 -> FOR.R2Prod (rr12FORRr1 rr11) (rr12FORRr1 rr12)
    Fm.R2Inv rr2 -> FOR.R2Inv (rr22FORRr2 rr2)
    Fm.R2Union rr21 rr22 -> FOR.R2Union (rr22FORRr2 rr21) (rr22FORRr2 rr22)
    Fm.R2Inter rr21 rr22 -> FOR.R2Inter (rr22FORRr2 rr21) (rr22FORRr2 rr22)
    Fm.R2Diff rr21 rr22 -> FOR.R2Diff (rr22FORRr2 rr21) (rr22FORRr2 rr22)
    Fm.R2Comp rr21 rr22 -> FOR.R2Comp (rr22FORRr2 rr21) (rr22FORRr2 rr22)

    Fm.R2Eq -> FOR.r2Eq
    Fm.R2Reflc rr2 -> FOR.r2Reflc (rr22FORRr2 rr2)
    Fm.R2Step trss -> FOR.r2Step (map trs2FORTrs trss)
    Fm.R2StepEq trss -> FOR.r2StepEq (map trs2FORTrs trss)
    Fm.R2Steps trss -> FOR.r2Steps (map trs2FORTrs trss)
    Fm.R2StepsEq trss -> FOR.r2StepsEq (map trs2FORTrs trss)
    Fm.R2StepsNF trss -> FOR.r2StepsNF (map trs2FORTrs trss)
    Fm.R2ParStep trss -> FOR.r2ParStep (map trs2FORTrs trss)
    Fm.R2RootStep trss -> FOR.r2RootStep (map trs2FORTrs trss)
    Fm.R2RootStepEq trss -> FOR.r2RootStepEq (map trs2FORTrs trss)
    Fm.R2RootSteps trss -> FOR.r2RootSteps (map trs2FORTrs trss)
    Fm.R2RootStepsEq trss -> FOR.r2RootStepsEq (map trs2FORTrs trss)
    Fm.R2NonRootStep trss -> FOR.r2NonRootStep (map trs2FORTrs trss)
    Fm.R2NonRootStepEq trss -> FOR.r2NonRootStepEq (map trs2FORTrs trss)
    Fm.R2NonRootSteps trss -> FOR.r2NonRootSteps (map trs2FORTrs trss)
    Fm.R2NonRootStepsEq trss -> FOR.r2NonRootStepsEq (map trs2FORTrs trss)
    Fm.R2Meet trss -> FOR.r2Meet (map trs2FORTrs trss)
    Fm.R2Join trss -> FOR.r2Join (map trs2FORTrs trss)

nat_of_int :: Int -> Nat
nat_of_int = nat_of_integer . fromIntegral

form2FORform :: Fm.Formula -> FOR.Formula Ftrs
form2FORform form = case form of
    Fm.FRR1 rr1 i -> FOR.FRR1 (rr12FORRr1 rr1) (nat_of_int i)
    Fm.FRR2 rr2 i j -> FOR.FRR2 (rr22FORRr2 rr2) (nat_of_int i) (nat_of_int j)
    Fm.FAnd fs -> FOR.FAnd (map form2FORform fs)
    Fm.FOr fs  -> FOR.FOr (map form2FORform fs)
    Fm.FNot f -> FOR.FNot (form2FORform f)
    Fm.FExists f -> FOR.FExists (form2FORform f)
    Fm.FForall f -> FOR.FForall (form2FORform f)
    Fm.FTrue -> FOR.fTrue
    Fm.FFalse -> FOR.fFalse
    Fm.FRestrict f is -> FOR.fRestrict (form2FORform f) (map nat_of_int is)
    Fm.FExtendSig f i ->
        error "Signature extension is not supported\n Only properties on ground terms allowed"

eq2FOREq :: Fm.Equivalence -> FOR.Equivalence
eq2FOREq Fm.EDistribAndOr = FOR.EDistribAndOr
eq2FOREq Fm.EDistribOrAnd = FOR.EDistribOrAnd

inf2FORInf :: Fm.Inference -> FOR.Inference Ftrs
inf2FORInf inf = case inf of
    Fm.IRR1 rr1 i -> FOR.IRR1 (rr12FORRr1 rr1) (nat_of_int i)
    Fm.IRR2 rr2 i j -> FOR.IRR2 (rr22FORRr2 rr2) (nat_of_int i) (nat_of_int j)
    Fm.IAnd is -> FOR.IAnd (map nat_of_int is)
    Fm.IOr is -> FOR.IOr (map nat_of_int is)
    Fm.INot i -> FOR.INot (nat_of_int i)
    Fm.IExists i -> FOR.IExists (nat_of_int i)
    Fm.IRename i is -> FOR.IRename (nat_of_int i) (map nat_of_int is)
    Fm.INNFPlus i -> FOR.INNFPlus (nat_of_int i)
    Fm.IRepl eq is i -> FOR.IRepl (eq2FOREq eq) (map nat_of_int is) (nat_of_int i)

claim2FORClaim :: Fm.Claim -> FOR.Claim
claim2FORClaim Fm.Empty = FOR.Emptya
claim2FORClaim Fm.Nonempty = FOR.Nonempty

info2FORinfo :: Fm.Info -> FOR.Info
info2FORinfo (Fm.Size i j k) = FOR.Size (nat_of_int i) (nat_of_int j) (nat_of_int k)

step2FORStep :: Fm.Step -> (Nat, (FOR.Inference Ftrs, (FOR.Formula Ftrs, [FOR.Info])))
step2FORStep (i, inf, form, info) =
    (nat_of_int i, (inf2FORInf inf,(form2FORform form, map info2FORinfo info)))

crt2FORCrt :: Fm.Certificate -> FOR.Certificate Ftrs
crt2FORCrt (Fm.Certificate ss cl i) =
    FOR.Certificate (map step2FORStep ss) (claim2FORClaim cl) (nat_of_int i)
